#include "ClothSpring.h"
#include "ClothNode.h"

FClothSpring::FClothSpring(const FClothNodeRef First, const FClothNodeRef Second, const float Elasticity,
	const float DampingFactor)
		: First(First), Second(Second), Elasticity(Elasticity), DampingFactor(DampingFactor)
{
	RestLength = FVector::Distance(First->GetPosition(), Second->GetPosition());
}

void FClothSpring::AddForceToNodes() const
{
	// We have two connected nodes - First and Second.
	// Compute the force acting between them as a result of this spring,
	// then add the force symmetrically to both nodes.
	FVector Force = GetActingForce();
	First->AddForce(Force);
	Second->AddForce(-Force);
}

FVector FClothSpring::GetActingForce() const
{
	// We need to compute the force with which this spring is acting.
	// For that purpose, we can use GetPosition and GetSpeed on each of our cloth nodes (First and Second)
	// and then do the calculation using the spring's parameters (fields DampingFactor, Elasticity and RestLength)
	FVector PositionDifference = Second->GetPosition() - First->GetPosition();
	FVector SpeedDifference = Second->GetSpeed() - First->GetSpeed();

	float Distance = PositionDifference.Size();

	// Damping force just pushes back in the opposite direction of the node's current speed (projected on the spring direction)
	FVector DampingForce = DampingFactor * FVector::DotProduct(SpeedDifference, PositionDifference) / Distance * PositionDifference.GetSafeNormal();
	// Elastic force pushes so that the spring returns to its resting length
	FVector ElasticForce = Elasticity * (PositionDifference - RestLength * PositionDifference.GetSafeNormal());

	return ElasticForce + DampingForce;
}
